@@ -44,7 +44,7 @@
  *	U3 V2.3         -	0xB3
  *	U3 Heavy V1.0   -	0x34
  *	U3 Heavy V1.1   -	0x35
- *      U3 Heavy V2.0   -       0x37
+ *  U3 Heavy V2.0   -   0x37
  *	U3 Light V1.0   -	0x38
  *	U3 Light V1.1   -	0x39
  *
@@ -73,6 +73,13 @@
 
 #define ALLOC_AGP_RANGE		0
 
+#define  NO_ATIR481_FASTWRITE	1
+
+#define kDevice_IDProperty			"device-id"
+#define kVendor_IDProperty			"vendor-id"
+#define kATIVendorID			0x00001002
+#define kATIR481DeviceID		0x00004a48
+
 #ifndef kIOAGPCommandValueKey
 #define kIOAGPCommandValueKey	"IOAGPCommandValue"
 #endif
@@ -89,6 +96,8 @@ OSDefineMetaClassAndStructors(AppleMacRiscAGP, AppleMacRiscPCI)
 
 OSDefineMetaClassAndStructors(AppleMacRiscHT, IOPCIBridge)
 
+OSDefineMetaClassAndStructors(AppleMacRiscPCIE, IOPCIBridge)
+
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
 
 bool AppleMacRiscPCI::start( IOService * provider )
@@ -143,6 +152,7 @@ bool AppleMacRiscPCI::start( IOService * provider )
             
     bridge = provider;
 
+    IODTSetResolving(provider, &compareAddressCell, &nvLocation);
     if( ! IODTResolveAddressCell( bridge, (UInt32 *) &ioAddrCell,
 		&ioPhys, &ioPhysLen) ) {
 
@@ -282,6 +292,9 @@ inline bool AppleMacRiscPCI::setConfigSpace( IOPCIAddressSpace space,
 {
     UInt32	addrCycle;
 
+    if( space.es.registerNumExtended)
+	return( false);
+
     offset &= 0xfc;
     if( space.s.busNum == secondaryBus) {
 
@@ -579,6 +592,9 @@ volatile UInt8 * AppleMacRiscHT::setConfigSpace( IOPCIAddressSpace space,
 {
     volatile UInt8 * configData;
 
+    if( space.es.registerNumExtended)
+	return( 0 );
+
     offset &= 0xfc;
     if( space.s.busNum == primaryBus) {
 
@@ -606,10 +622,14 @@ UInt32 AppleMacRiscHT::configRead32( IOPCIAddressSpace space,
 
     configData = setConfigSpace( space, offset );
 
-    offset = offset & 3 & 4;
-
-    data = OSReadSwapInt32( configData, offset );
-    eieio();
+    if (configData)
+    {
+	offset = offset & 3 & 4;
+	data = OSReadSwapInt32( configData, offset );
+	eieio();
+    }
+    else
+	data = 0xffffffff;
 
     return( data );
 }
@@ -621,6 +641,303 @@ void AppleMacRiscHT::configWrite32( IOPCIAddressSpace space,
 
     configData = setConfigSpace( space, offset );
 
+    if (configData)
+    {
+	offset = offset & 3 & 4;
+
+	OSWriteSwapInt32( configData, offset, data );
+	eieio();
+	/* read to sync */
+	(void) OSReadSwapInt32( configData, offset );
+	eieio();
+	sync();
+	isync();
+    }
+}
+
+UInt16 AppleMacRiscHT::configRead16( IOPCIAddressSpace space,
+					UInt8 offset )
+{
+    volatile UInt8 * configData;
+    UInt16	     data;
+
+    configData = setConfigSpace( space, offset );
+
+    if (configData)
+    {
+	offset = offset & 3 & 6;
+
+	data = OSReadSwapInt16( configData, offset );
+	eieio();
+    }
+    else
+	data = 0xffff;
+
+    return( data );
+}
+
+void AppleMacRiscHT::configWrite16( IOPCIAddressSpace space, 
+					UInt8 offset, UInt16 data )
+{
+    volatile UInt8 * configData;
+
+    configData = setConfigSpace( space, offset );
+
+    if (configData)
+    {
+	offset = offset & 3 & 6;
+
+	OSWriteSwapInt16( configData, offset, data );
+	eieio();
+	/* read to sync */
+	(void) OSReadSwapInt16( configData, offset );
+	eieio();
+	sync();
+	isync();
+    }
+}
+
+UInt8 AppleMacRiscHT::configRead8( IOPCIAddressSpace space,
+					UInt8 offset )
+{
+    volatile UInt8 * configData;
+    UInt16	     data;
+
+    configData = setConfigSpace( space, offset );
+
+    if (configData)
+    {
+	offset = offset & 3;
+	data = configData[ offset ];
+	eieio();
+    }
+    else
+	data = 0xff;
+
+    return( data );
+}
+
+void AppleMacRiscHT::configWrite8( IOPCIAddressSpace space, 
+					UInt8 offset, UInt8 data )
+{
+    volatile UInt8 * configData;
+
+    configData = setConfigSpace( space, offset );
+
+    if (configData)
+    {
+	offset = offset & 3;
+
+	configData[ offset ] = data;
+	eieio();
+	/* read to sync */
+	data = configData[ offset ];
+	eieio();
+	sync();
+	isync();
+    }
+}
+
+/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
+
+bool AppleMacRiscPCIE::start( IOService * provider )
+{
+    IOPhysicalAddress		ioPhysAddrs;
+    IOPhysicalAddress		ioPhysLengths;
+    OSArray *			array;
+    IODeviceMemory::InitElement	rangeList[ 2 ];
+    IODeviceMemory *		md;
+    IORegistryEntry *		bridge;
+
+    bridge = provider;
+
+    md = provider->getDeviceMemoryWithIndex(0);
+    if (md == 0)
+	return 0;
+    ioPhysAddrs = md->getPhysicalAddress();
+    ioPhysLengths = md->getLength();
+
+    if( 0 == (lock = IOSimpleLockAlloc()))
+	return( false );
+
+    /* define more explicit ranges */
+
+    rangeList[0].start	= ioPhysAddrs;
+    rangeList[0].length = 0x01000000UL;
+    rangeList[1].start	= ioPhysAddrs + 0x01000000;
+    rangeList[1].length	= 0x01000000UL;
+
+    IORangeAllocator * platformRanges;
+    platformRanges = IOService::getPlatform()->getPhysicalRangeAllocator();
+    assert( platformRanges );
+    platformRanges->allocateRange( ioPhysAddrs, ioPhysLengths );
+
+    array = IODeviceMemory::arrayFromList( rangeList, 2 );
+    if( !array)
+	return( false);
+
+    provider->setDeviceMemory( array );
+    array->release();
+
+    /* map registers */
+
+    if( (hostRegsMap = provider->mapDeviceMemoryWithIndex( 0 )))
+         hostRegs = (volatile UInt8 *) hostRegsMap->getVirtualAddress();
+    if( (configAtomicMap = provider->mapDeviceMemoryWithIndex( 1 )))
+        configAtomic = (volatile UInt8 *) configAtomicMap->getVirtualAddress();
+
+    if( !hostRegs || !configAtomic)
+	return( false);
+
+    configAddrNA = (volatile UInt32 *)(hostRegs + 0x800000);
+    configDataNA = (volatile UInt8 *)(hostRegs + 0xC00000);
+
+    // Primary Bus is always zero.
+    primaryBus = 0;
+
+    return( super::start( provider));
+}
+
+bool AppleMacRiscPCIE::configure( IOService * provider )
+{
+    UInt32	addressSelects;
+    UInt32	index;
+    bool	ok;
+
+    addressSelects = configRead32( getBridgeSpace(), kMacRISCPCIEAddressSelect );
+
+    coarseAddressMask	= addressSelects >> 16;
+    fineAddressMask	= addressSelects & 0xffff;
+
+    for( index = 0; index < 15; index++ ) {
+		if( coarseAddressMask & (1 << index)) {
+			ok = addBridgeMemoryRange( index << 28, 0x10000000, true );
+		}
+    }
+
+//    if( coarseAddressMask & (1 << 15))	// F segment
+	for( index = 0; index < 15; index++ ) {
+		if( fineAddressMask & (1 << index)) {
+		ok = addBridgeMemoryRange( (0xf0 | index) << 24,
+						0x01000000, true );
+	    }
+	}
+
+    return( super::configure( provider));
+}
+
+void AppleMacRiscPCIE::free()
+{
+    if( hostRegsMap)
+	hostRegsMap->release();
+    if( configAtomicMap)
+	configAtomicMap->release();
+    if( lock)
+	IOSimpleLockFree( lock);
+
+    super::free();
+}
+
+IODeviceMemory * AppleMacRiscPCIE::ioDeviceMemory( void )
+{
+    return( 0 );
+}
+
+IOPCIAddressSpace AppleMacRiscPCIE::getBridgeSpace( void )
+{
+    IOPCIAddressSpace	space;
+
+    space.bits = 0;
+    space.es.busNum = primaryBus;
+    space.es.deviceNum = kPCIEBridgeSelfDevice;
+
+    return( space );
+}
+
+bool AppleMacRiscPCIE::configCycleAtomic( IOPCIAddressSpace space)
+{
+  if( space.es.registerNumExtended != 0)
+    return false;
+
+  return true;
+}
+
+volatile UInt8 * AppleMacRiscPCIE::setConfigSpace( IOPCIAddressSpace space,
+                                            UInt8 offset )
+{
+    UInt32	addrCycle;
+    volatile UInt8 * configData;
+
+    offset &= 0xfc;
+
+    if( configCycleAtomic(space)) {
+      configData = configAtomic + ((space.bits & 0x00ffff00) | offset);
+
+    } else {
+
+      if( space.es.busNum == primaryBus) {
+        // primary config cycle
+        addrCycle = (	  (1 << (space.es.deviceNum + 11))
+						| (space.es.registerNumExtended << 28)
+						| (space.es.functionNum << 8)
+						| offset );
+
+      } else {
+		// pass thru config cycle
+		addrCycle = (	  ((space.bits) & 0x00ffff00)
+						| (space.es.registerNumExtended << 28)
+						| offset
+						| 1 );
+      }
+
+      do {
+        OSWriteSwapInt32( configAddrNA, 0, addrCycle);
+        eieio();
+      } while( addrCycle != OSReadSwapInt32( configAddrNA, 0 ));
+      eieio();
+
+      configData = configDataNA;
+    }
+
+    return( configData );
+}
+
+UInt32 AppleMacRiscPCIE::configRead32( IOPCIAddressSpace space,
+					UInt8 offset )
+{
+    volatile UInt8 * configData;
+    UInt32	     data;
+    bool	     atomic = configCycleAtomic(space);
+    IOInterruptState ints = 0;
+
+    if( !atomic)
+      ints = IOSimpleLockLockDisableInterrupt( lock );
+
+    configData = setConfigSpace( space, offset );
+
+    offset = offset & 3 & 4;
+
+    data = OSReadSwapInt32( configData, offset );
+    eieio();
+
+    if( !atomic)
+      IOSimpleLockUnlockEnableInterrupt( lock, ints );
+
+    return( data );
+}
+
+void AppleMacRiscPCIE::configWrite32( IOPCIAddressSpace space, 
+					UInt8 offset, UInt32 data )
+{
+    volatile UInt8 * configData;
+    bool	     atomic = configCycleAtomic(space);
+    IOInterruptState ints = 0;
+
+    if( !atomic)
+      ints = IOSimpleLockLockDisableInterrupt( lock );
+
+    configData = setConfigSpace( space, offset );
+
     offset = offset & 3 & 4;
 
     OSWriteSwapInt32( configData, offset, data );
@@ -630,13 +947,21 @@ void AppleMacRiscHT::configWrite32( IOPCIAddressSpace space,
     eieio();
     sync();
     isync();
+
+    if( !atomic)
+      IOSimpleLockUnlockEnableInterrupt( lock, ints );
 }
 
-UInt16 AppleMacRiscHT::configRead16( IOPCIAddressSpace space,
+UInt16 AppleMacRiscPCIE::configRead16( IOPCIAddressSpace space,
 					UInt8 offset )
 {
     volatile UInt8 * configData;
     UInt16	     data;
+    bool	     atomic = configCycleAtomic(space);
+    IOInterruptState ints = 0;
+
+    if( !atomic)
+      ints = IOSimpleLockLockDisableInterrupt( lock );
 
     configData = setConfigSpace( space, offset );
 
@@ -645,13 +970,21 @@ UInt16 AppleMacRiscHT::configRead16( IOPCIAddressSpace space,
     data = OSReadSwapInt16( configData, offset );
     eieio();
 
+    if( !atomic)
+      IOSimpleLockUnlockEnableInterrupt( lock, ints );
+
     return( data );
 }
 
-void AppleMacRiscHT::configWrite16( IOPCIAddressSpace space, 
+void AppleMacRiscPCIE::configWrite16( IOPCIAddressSpace space, 
 					UInt8 offset, UInt16 data )
 {
     volatile UInt8 * configData;
+    bool	     atomic = configCycleAtomic(space);
+    IOInterruptState ints = 0;
+
+    if( !atomic)
+      ints = IOSimpleLockLockDisableInterrupt( lock );
 
     configData = setConfigSpace( space, offset );
 
@@ -664,13 +997,21 @@ void AppleMacRiscHT::configWrite16( IOPCIAddressSpace space,
     eieio();
     sync();
     isync();
+
+    if( !atomic)
+      IOSimpleLockUnlockEnableInterrupt( lock, ints );
 }
 
-UInt8 AppleMacRiscHT::configRead8( IOPCIAddressSpace space,
+UInt8 AppleMacRiscPCIE::configRead8( IOPCIAddressSpace space,
 					UInt8 offset )
 {
     volatile UInt8 * configData;
     UInt16	     data;
+    bool	     atomic = configCycleAtomic(space);
+    IOInterruptState ints = 0;
+
+    if ( !atomic)
+      ints = IOSimpleLockLockDisableInterrupt( lock );
 
     configData = setConfigSpace( space, offset );
 
@@ -679,13 +1020,21 @@ UInt8 AppleMacRiscHT::configRead8( IOPCIAddressSpace space,
     data = configData[ offset ];
     eieio();
 
+    if( !atomic)
+      IOSimpleLockUnlockEnableInterrupt( lock, ints );
+
     return( data );
 }
 
-void AppleMacRiscHT::configWrite8( IOPCIAddressSpace space, 
+void AppleMacRiscPCIE::configWrite8( IOPCIAddressSpace space, 
 					UInt8 offset, UInt8 data )
 {
     volatile UInt8 * configData;
+    bool	     atomic = configCycleAtomic(space);
+    IOInterruptState ints = 0;
+
+    if( !atomic)
+      ints = IOSimpleLockLockDisableInterrupt( lock );
 
     configData = setConfigSpace( space, offset );
 
@@ -698,6 +1047,90 @@ void AppleMacRiscHT::configWrite8( IOPCIAddressSpace space,
     eieio();
     sync();
     isync();
+
+    if( !atomic)
+      IOSimpleLockUnlockEnableInterrupt( lock, ints );
+}
+
+// whatToDo for setDevicePowerState()
+enum
+{
+    kSaveDeviceState    = 0,
+    kRestoreDeviceState = 1,
+    kSaveBridgeState    = 2,
+    kRestoreBridgeState = 3
+};
+
+
+IOReturn AppleMacRiscPCIE::setDevicePowerState( IOPCIDevice * device, unsigned long whatToDo )
+{
+	if (kSaveBridgeState == whatToDo)
+	{
+		saveBridgeState();
+	}
+	else if (kRestoreBridgeState == whatToDo)
+	{
+		restoreBridgeState();
+	}
+	
+	return super::setDevicePowerState(device, whatToDo);
+}
+
+void AppleMacRiscPCIE::saveBridgeState( void )
+{
+    UInt32 cnt;
+
+    for (cnt = 0; cnt < kMacRISCPCIEBridgeRegs; cnt++)
+    {
+        bridgeState[cnt] = configRead32(getBridgeSpace(), cnt * 4);
+    }
+}
+
+void AppleMacRiscPCIE::restoreBridgeState( void )
+{
+	UInt32				cnt;
+	UInt32				addrCycle;
+    IOPCIAddressSpace	space = getBridgeSpace();
+    IOInterruptState ints = 0;
+        
+    // Start by restoring the PCIE primary/secondary bus numbers so that we can 
+    // can do direct access config cycles.
+    
+    ints = IOSimpleLockLockDisableInterrupt( lock );
+
+    addrCycle = ( (1 << (space.es.deviceNum + 11))
+                      | (space.es.registerNumExtended << 28)
+                      | (space.es.functionNum << 8)
+                      | 18 );
+    do {
+        OSWriteSwapInt32( configAddrNA, 0, addrCycle);
+        eieio();
+      } while( addrCycle != OSReadSwapInt32( configAddrNA, 0 ));
+      eieio();
+
+    OSWriteSwapInt32( configDataNA, 18,  bridgeState[18] );
+    eieio();
+
+    IOSimpleLockUnlockEnableInterrupt( lock, ints );
+
+  
+    // start at config space location 8 -- bytes 0-3 are
+    // defined by the PCI Spec. as ReadOnly, and we don't
+    // want to write anything to the Command or Status
+    // registers until the rest of config space is set up.
+
+    for (cnt = (kIOPCIConfigCommand >> 2) + 1; cnt < kMacRISCPCIEBridgeRegs; cnt++)
+    {
+        configWrite32(space, cnt * 4, bridgeState[cnt]);
+    }
+
+    // once the rest of the config space is restored,
+    // turn on all the enables (,etc.) in the Command register.
+    // NOTE - we also reset any status bits in the Status register
+    // that may have been previously indicated by writing a '1'
+    // to the bits indicating whatever they were indicating.
+
+    configWrite32(space, kIOPCIConfigCommand, bridgeState[kIOPCIConfigCommand >> 2]);
 }
 
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
@@ -870,6 +1303,21 @@ IOReturn AppleMacRiscAGP::createAGPSpace( IOAGPDevice * master,
 //    agpCommandMask &= ~kIOAGPFastWrite;
 //#endif	/* !NO_FASTWRITE */
 
+#if NO_ATIR481_FASTWRITE
+
+	const UInt32    atiID = kATIVendorID;
+	const UInt32    deviceR481ID = kATIR481DeviceID;
+    OSData *		data1;
+	
+	 if (  (data = OSDynamicCast(OSData, master->getProperty(kVendor_IDProperty))) 
+		&& (data->isEqualTo( &atiID, sizeof(atiID))) 
+		&& (data1 = OSDynamicCast(OSData, master->getProperty(kDevice_IDProperty))) 
+	    && (data1->isEqualTo( &deviceR481ID, sizeof(deviceR481ID))) )
+        {
+			agpCommandMask &= ~kIOAGPFastWrite;			// ATI R481 - disable Fast writes for R481
+		}
+#endif
+
     if ((data = OSDynamicCast(OSData, getProvider()->getProperty("AAPL,agp-clear"))))
 	agpCommandMask &= ~*((UInt32 *)data->getBytesNoCopy());
     if ((data = OSDynamicCast(OSData, getProvider()->getProperty("AAPL,agp-set"))))
